import torch
from torch import nn

from spinesTS.layers import Time2Vec, MoveAvg
from spinesTS.base import TorchModelMixin, ForecastingMixin


class T2V(nn.Module):
    def __init__(self, in_shapes, out_features, flip_features=False, ma_window_size=3):
        super(T2V, self).__init__()
        assert isinstance(ma_window_size, int)
        self.in_shape_type = type(in_shapes)
        if self.in_shape_type == tuple:
            rows, self.in_features = in_shapes
        else:
            self.in_features, self.out_features = in_shapes, out_features

        if ma_window_size > 0:
            self.in_features = self.in_features - ma_window_size + 2
            self.move_avg = MoveAvg(kernel_size=ma_window_size)
        else:
            self.move_avg = lambda s: s

        self.t2v = Time2Vec(self.in_features)
        if flip_features:
            self.t2v2 = Time2Vec(self.in_features)
            ln_layer_in_fea = 4 * self.in_features
        else:
            ln_layer_in_fea = 2 * self.in_features

        if self.in_shape_type == tuple:
            self.linear = nn.Linear(ln_layer_in_fea * rows, out_features)
        else:
            self.linear = nn.Linear(ln_layer_in_fea, out_features)

        self.flip_features = flip_features

    def forward(self, x):
        x = self.move_avg(x)
        x1 = self.t2v(x)
        if self.flip_features:
            _x = x.clone()
            x2 = self.t2v2(torch.flip(_x, dims=[-1]))
            x = torch.concat((x1, x2), dim=-1)
        else:
            x = x1

        if self.in_shape_type == tuple:
            return self.linear(x.reshape(-1, x.shape[1] * x.shape[2]))
        else:
            return self.linear(x)


class Time2VecNet(TorchModelMixin, ForecastingMixin):
    def __init__(self, in_features, out_features, flip_features=False, learning_rate=0.01,
                 random_seed=42, device=None, ma_window_size=3, loss_fn='mae'):
        super(Time2VecNet, self).__init__(random_seed, device=device, loss_fn=loss_fn)
        self.in_features, self.out_features = in_features, out_features
        self.learning_rate = learning_rate
        self.flip_features = flip_features
        self.ma_window_size = ma_window_size
        self.model, self.loss_fn, self.optimizer = self.call()

    def call(self):
        model = T2V(self.in_features, self.out_features, flip_features=self.flip_features,
                    ma_window_size=self.ma_window_size)
        loss_fn = self.loss_fn
        optimizer = torch.optim.AdamW(model.parameters(), lr=self.learning_rate)
        return model, loss_fn, optimizer

    def fit(
            self,
            X_train,
            y_train,
            epochs=1000,
            batch_size='auto',
            eval_set=None,
            monitor='val_loss',
            min_delta=0,
            patience=10,
            lr_scheduler='ReduceLROnPlateau',
            lr_scheduler_patience=10,
            lr_factor=0.7,
            restore_best_weights=True,
            verbose=True,
            **kwargs
    ):
        return super().fit(X_train, y_train, epochs, batch_size, eval_set, loss_type='down', metrics_name='mae',
                           monitor=monitor, lr_scheduler=lr_scheduler,
                           lr_scheduler_patience=lr_scheduler_patience,
                           lr_factor=lr_factor,
                           min_delta=min_delta, patience=patience, restore_best_weights=restore_best_weights,
                           verbose=verbose, **kwargs)



import math

import torch
import torch.nn as nn


class PositionalEncoding(nn.Module):
    def __init__(self, d_model, max_len=5000, add_x=True):
        super(PositionalEncoding, self).__init__()
        # Compute the positional encodings once in log space.
        self.add_x = add_x
        pe = torch.zeros(max_len, d_model).float()
        pe.require_grad = False

        position = torch.arange(0, max_len).float().unsqueeze(1)
        div_term = (torch.arange(0, d_model, 2).float()
                    * -(math.log(10000.0) / d_model)).exp()

        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)

        pe = pe.unsqueeze(0)
        self.register_buffer('pe', pe)

    def forward(self, x):
        if self.add_x:
            return x + self.pe[:, :x.shape[1]]
        return self.pe[:, :x.shape[1]]





from functools import partial
from typing import Union, Any

import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split


def split_series(x_seq, y_seq, window_size: int, pred_steps: int, train_size: Union[None, float] = None,
                 shuffle: bool = False, skip_steps: int = 1):
    """Returns two-dimensional array cut by the specified window size.

    Parameters
    ----------
    x_seq : numpy.ndarray or pandas.Series or list, the series which needed to split
    y_seq : numpy.ndarray or pandas.Series or list, the series which needed to split
    window_size : int, sliding window size
    pred_steps : int, the number of steps predicted forward by the series
    train_size : float,
    shuffle : bool, whether to shuffle the split results
    skip_steps : int, the number of skipped steps per slide

    Returns
    -------
    numpy.ndarray.

    when train_size is not None, return X_train, X_test, y_train, y_test
    otherwise return X, y
    """
    assert isinstance(x_seq, (pd.Series, np.ndarray, list)) and isinstance(y_seq, (pd.Series, np.ndarray, list))
    assert train_size is None or (0 < train_size < 1 and isinstance(train_size, float))
    assert isinstance(shuffle, bool)
    assert isinstance(pred_steps, int) and pred_steps > 0

    if isinstance(x_seq, pd.Series):
        x_seq = x_seq.values
    elif isinstance(x_seq, list):
        x_seq = np.array(x_seq)
    if isinstance(y_seq, pd.Series):
        y_seq = y_seq.values
    elif isinstance(y_seq, list):
        y_seq = np.array(y_seq)
    X, y = [], []

    for i in range(len(x_seq)):
        end_index = i * skip_steps + window_size
        out_end_index = end_index + pred_steps

        if out_end_index > len(x_seq):
            break

        seq_x, seq_y = list(x_seq[i * skip_steps:end_index]), list(y_seq[end_index:out_end_index])

        X.append(seq_x)
        y.append(seq_y)

    if train_size is None:
        if pred_steps > 1:
            return np.array(X), np.array(y)
        return np.array(X), np.squeeze(np.array(y))
    elif pred_steps == 1 and train_size is not None:
        return train_test_split(np.array(X), np.squeeze(np.array(y)), train_size=train_size, shuffle=shuffle)
    else:
        return train_test_split(np.array(X), np.array(y), train_size=train_size, shuffle=shuffle)


def lag_splits(x_seq, window_size, skip_steps=1, pred_steps=1):
    """Returns a two-dimensional array cut by the specified window size.

    Parameters
    ----------
    x_seq : numpy.ndarray or pandas.Series or list, the series which needed to split
    window_size : int, sliding window size
    skip_steps : int, the number of skipped steps per slide
    pred_steps : int, the number of steps predicted forward by the series

    Return
    -------
    numpy.ndarray
    """
    assert isinstance(x_seq, (pd.Series, np.ndarray, list))

    if isinstance(x_seq, pd.Series):
        x_seq = x_seq.values
    elif isinstance(x_seq, list):
        x_seq = np.array(x_seq)

    X = []
    p, _ = 0, 0
    for i in range(len(x_seq)):
        if p == pred_steps:
            p, _ = 1, i
        else:
            p += 1
        end_index = _ * skip_steps + window_size
        if end_index > len(x_seq):
            break

        seq_x = list(x_seq[_ * skip_steps:end_index])

        X.append(seq_x)

    return np.array(X)


train_test_split_ts = partial(train_test_split, shuffle=False)  # train_test_split for time series



import numpy as np
import pandas as pd

from spinesTS.base import ForecastingMixin
from spinesTS.ml_model import MultiOutputRegressor
from spinesTS.utils import check_is_fitted
from spinesTS.pipeline import Pipeline
from spinesTS.preprocessing import split_series, lag_splits, moving_average
from spinesTS.features_generator import DataExtendFeatures


class GBRTPreprocessing:
    """WideGBRT features engineering class"""
    def __init__(self, in_features, out_features, target_col,
                 train_size=0.8, date_col=None, differential_n=0, moving_avg_n=0,
                 extend_daily_target_features=True
                 ):
        self.cf = None
        self.input_features = in_features
        self.output_features = out_features
        self.target_col = target_col
        self.train_size = train_size
        self.date_col = date_col
        self.extend_daily_target_features = extend_daily_target_features
        assert isinstance(differential_n, int) and differential_n >= 0
        self.differential_n = differential_n
        assert isinstance(moving_avg_n, int) and moving_avg_n >= 0
        self.moving_avg_n = moving_avg_n
        self.x_shape = None

        self.__spinesTS_is_fitted__ = False

    def process_date_col(self, x):
        """Processing date column"""
        return DataExtendFeatures(date_col=self.date_col, drop_date_col=True).fit_transform(x)

    @staticmethod
    def process_target_col(x):
        assert x.ndim == 2

        mean_res = x.mean(axis=1).reshape((-1, 1))
        median_res = np.percentile(x, q=50, axis=1).reshape((-1, 1))
        min_res = x.min(axis=1).reshape((-1, 1))
        max_res = x.max(axis=1).reshape((-1, 1))
        p25 = np.percentile(x, q=25, axis=1).reshape((-1, 1))
        p75 = np.percentile(x, q=75, axis=1).reshape((-1, 1))
        std = np.std(x, axis=1).reshape((-1, 1))

        return np.concatenate((mean_res, median_res, max_res, min_res, p25, p75, std), axis=1)

    def check_x_types(self, x):
        assert isinstance(x, (pd.DataFrame, np.ndarray))

        if not isinstance(x, pd.DataFrame):
            assert x.ndim == 2, "Only accept two-dim numpy.ndarray."
            if not isinstance(self.target_col, int):
                raise TypeError("when `x` is of type `numpy.ndarray`, the `target_col` parameter must be an integer.")

            if self.date_col is not None and not isinstance(self.date_col, int):
                raise TypeError("when `x` is of type `numpy.ndarray`, the `date_col` parameter must be an integer.")

    def fit(self, x):
        self.check_x_types(x)
        self.x_shape = x.shape[1]
        self.__spinesTS_is_fitted__ = True
        return self

    def transform(self, x, mode='train'):
        """Transform data to fit WideGBRT model.

        result's columns sequence:
        lag_1, lag_2, lag_3, ..., lag_n, x_col_1, x_col_2, ..., x_col_n, date_fea_1, date_fea_2, ..., date_fea_n

        Parameters
        ----------
        x: spines.data.DataTS or pandas.core.DataFrame or numpy.ndarray, the data that needs to be transformed
        mode: ('train', 'predict'), the way to transform data, default: 'train'

        Returns
        -------
        numpy.ndarray, x_train, x_test, y_train, y_test, when mode = 'train', else, x, y

        """
        assert mode in ('train', 'predict')
        check_is_fitted(self)

        self.check_x_types(x)

        if x.shape[1] != self.x_shape:
            raise ValueError("data shape does not match the shape of the data at the time of fitting.")

        _tar = x[self.target_col].values if isinstance(x, pd.DataFrame) else x[:, self.target_col]

        if isinstance(x, pd.DataFrame):
            if self.date_col is not None:
                x = self.process_date_col(x)
            # timestamp features
            _non_lag_fea = x.loc[:, [i for i in x.columns if i != self.target_col]].values
        else:
            if self.date_col is not None:
                x = self.process_date_col(pd.DataFrame(x, columns=range(x.shape[1]))).values

            _non_lag_fea = x[:, [i for i in range(x.shape[1]) if i != self.target_col]]

        if mode == 'train':
            if self.train_size is None:
                x, y = split_series(_tar, _tar, self.input_features, self.output_features, train_size=self.train_size)

                if self.extend_daily_target_features:
                    tar_fea_x = self.process_target_col(x)

                if self.moving_avg_n > 0:
                    x = moving_average(x, window_size=self.moving_avg_n)

                if self.differential_n > 0:
                    x = np.diff(x, axis=1, n=self.differential_n)

                x_non_lag, _ = split_series(_non_lag_fea, _tar, self.input_features,
                                            self.output_features, train_size=self.train_size)

                if x_non_lag.shape[1] > 0:
                    if self.extend_daily_target_features:
                        return np.concatenate((x, tar_fea_x, self._process_x_non_lag_dim(x_non_lag)), axis=1), y
                    return np.concatenate((x, self._process_x_non_lag_dim(x_non_lag)), axis=1), y
                else:
                    if self.extend_daily_target_features:
                        return np.concatenate((x, tar_fea_x), axis=1), y
                    return x, y
            else:
                x_train, x_test, y_train, y_test = split_series(_tar, _tar, self.input_features,
                                                                self.output_features, train_size=self.train_size)

                if self.extend_daily_target_features:
                    tar_fea_x_train = self.process_target_col(x_train)
                    tar_fea_x_test = self.process_target_col(x_test)

                if self.moving_avg_n > 0:
                    x_train = moving_average(x_train, window_size=self.moving_avg_n)
                    x_test = moving_average(x_test, window_size=self.moving_avg_n)

                if self.differential_n > 0:
                    x_train = np.diff(x_train, axis=1, n=self.differential_n)
                    x_test = np.diff(x_test, axis=1, n=self.differential_n)

                x_non_lag_train, x_non_lag_test, _, _ = split_series(_non_lag_fea, _tar, self.input_features,
                                                                     self.output_features, train_size=self.train_size)
                # columns sequence:
                # lag_1, lag_2, lag_3, ..., lag_n, x_col_1, x_col_2, ..., x_col_n,
                # date_fea_1, date_fea_2, ..., date_fea_n
                if len(x_non_lag_train) > 0 and len(x_non_lag_test) > 0:
                    if self.extend_daily_target_features:
                        return np.concatenate((x_train, tar_fea_x_train, self._process_x_non_lag_dim(x_non_lag_train)), axis=1), \
                            np.concatenate((x_test, tar_fea_x_test, self._process_x_non_lag_dim(x_non_lag_test)), axis=1), y_train, y_test
                    return np.concatenate((x_train, self._process_x_non_lag_dim(x_non_lag_train)), axis=1), \
                        np.concatenate((x_test, self._process_x_non_lag_dim(x_non_lag_test)), axis=1), y_train, y_test
                else:
                    if self.extend_daily_target_features:
                        np.concatenate((x_train, tar_fea_x_train), axis=1), \
                            np.concatenate((x_test, tar_fea_x_test), axis=1), y_train, y_test
                    return x_train, x_test, y_train, y_test
        else:
            split_tar = lag_splits(
                _tar, window_size=self.input_features, skip_steps=1, pred_steps=1
            )[:-self.output_features]

            if self.extend_daily_target_features:
                tar_fea_x = self.process_target_col(split_tar)

            if self.moving_avg_n > 0:
                split_tar = moving_average(split_tar, window_size=self.moving_avg_n)

            if self.differential_n > 0:
                split_tar = np.diff(split_tar, axis=1, n=self.differential_n)

            split_non_lag_fea = lag_splits(
                _non_lag_fea, window_size=self.input_features, skip_steps=1, pred_steps=1
            )[:-self.output_features]

            if len(split_non_lag_fea) > 0:
                if self.extend_daily_target_features:
                    return np.concatenate((split_tar, tar_fea_x, self._process_x_non_lag_dim(split_non_lag_fea)), axis=1)
                return np.concatenate((split_tar, self._process_x_non_lag_dim(split_non_lag_fea)), axis=1)
            else:
                if self.extend_daily_target_features:
                    return np.concatenate((split_tar, tar_fea_x), axis=1)
                return split_tar

    @staticmethod
    def _process_x_non_lag_dim(x):
        if x[:, -1, :].squeeze().ndim == 1 and x[:, -1, :].ndim == 2:
            return x[:, -1, :]
        elif x[:, -1, :].squeeze().ndim == 1:
            return x[:, -1, :].squeeze(1)
        return x[:, -1, :].squeeze()

class WideGBRT(ForecastingMixin):
    def __init__(self, model, scaler=None, is_pipeline=False):
        """
        
        Parameters
        -----------
        model: estimator, note that estimator should implement the fit and predict method.
        scaler: Numerical scale normalizer, note that scaler should implement the fit and predict method.
            Default to None.
        is_pipeline: whether model is a sklearn-type pipeline,
            if not, class WideGBRT will automatically assemble a pipeline to predict multiple target values.
            Default to False.

        Returns
        -------
        None

        """
        if not is_pipeline:
            if scaler:
                multi_reg = Pipeline([
                    ('sc', scaler),
                    ('multi_reg', MultiOutputRegressor(model))
                ])
    
                self.model = multi_reg
            else:
                self.model = MultiOutputRegressor(model)
        else:
            self.model = model

        self.__spinesTS_is_fitted__ = False

    def fit(self, X, y, **model_fit_kwargs):
        self.model.fit(X, y, **model_fit_kwargs)

        self.__spinesTS_is_fitted__ = True
        return self

    def predict(self, X, **model_predict_kwargs):
        check_is_fitted(self)
        return self.model.predict(X, **model_predict_kwargs)



from typing import *
import copy

import numpy as np


class DataGenerator:
    """Generate timeseries-like data
    """

    @staticmethod
    def trigonometry_ds(size: int = 1000, sin_cos_noise_fact: Sized = (0.5, 0.3, 0.2),
                        random_state: Optional[int] = None):
        """Generates a weighted combination sequence of sine and cosine waves

        Parameters
        ----------
        size : int, data points number to generate
        sin_cos_noise_fact : Tuple of float, the weights of sine, cosine and standard gaussian distributed noise
        random_state : None or int, random seed

        Returns
        -------
        numpy.ndarray
        """
        assert size is not None and isinstance(size, (int, float)) is True
        assert random_state is None or isinstance(random_state, int) is True
        assert len(sin_cos_noise_fact) == 3
        np.random.seed(random_state)
        s = int(np.ceil(size))

        assert s > 0

        ds = np.zeros(s)
        ds = ds + np.array([np.sin(i) for i in range(s)]) * sin_cos_noise_fact[0] + \
             np.array([np.cos(i) for i in range(s)]) * sin_cos_noise_fact[1] + \
             np.array([np.random.randn() for i in range(s)]) * sin_cos_noise_fact[2]

        return ds

    @staticmethod
    def white_noise(size=1000, mean=0., std=1., random_state=None):
        """Generates a sequence of white noise(gaussian distribution noise)

        Parameters
        ----------
        size : int, data points number to generate
        mean : float, the mean of gaussian distribution noise, default to 0.
        std : float, the standard deviation (std) of gaussian distribution noise, default to 1.
        random_state : None or int, random seed

        Returns
        -------
        numpy.ndarray
        """
        assert size is not None and isinstance(size, (int, float)) is True
        assert random_state is None or isinstance(random_state, int) is True
        assert isinstance(mean, float) is True and isinstance(std, float) is True

        np.random.seed(random_state)
        s = int(np.ceil(size))

        assert s > 0

        ds = np.random.normal(mean, std, size=s)

        return ds

    @staticmethod
    def random_walk(size=1000, started_zero=True, random_state=None):
        """Generates a sequence of random walk data point

        Parameters
        ----------
        size : int, data points number to generate
        started_zero : bool, whether to start from zero
        random_state : None or int, random seed

        Returns
        -------
        numpy.ndarray
        """
        assert size is not None and isinstance(size, (int, float))
        assert random_state is None or isinstance(random_state, int)
        assert isinstance(started_zero, bool) is True

        np.random.seed(random_state)
        s = int(np.ceil(size))

        if started_zero is True:
            start_p = 0.
        else:
            start_p = np.random.randn()

        res = [start_p]

        while len(res) < s:
            _ = copy.deepcopy(res[-1])
            mid = copy.deepcopy(_)
            _ += np.random.randn()
            if _ != mid:
                res.append(_)

        ds = np.array(res)

        return ds


class RandomEventGenerator:
    """Generate a random time series with true seasonal and temporal trends,
        intermingled with Gaussian noise

    Parameters
    ----------
    size : int, number of time collection points.
    seasons : int, seasonal cycle length.
    random_state : Random seed.

    Returns
    -------
    None
    """
    def __init__(self, size=10000, seasons=None, sin_cos_noise_fact=(0.5, 0.3, 0.2), stacking_level=6, random_state=None):
        assert isinstance(stacking_level, int)
        self.seasons = seasons
        self.size = size
        self.random_state = random_state
        self.sin_cos_noise_fact = sin_cos_noise_fact
        self.stacking_level = stacking_level

    def _seasonal(self):
        """Get seasonal trends data.

        Returns
        -------
        numpy.ndarray
        """
        np.random.seed(self.random_state)
        s = np.random.randn(self.seasons)
        return np.concatenate(
            (np.repeat(s, self.size // self.seasons),
             s[:self.size % self.seasons] if self.size % self.seasons != 0 else []))

    def _noise(self):
        """Get Gaussian noise.

        Returns
        -------
        numpy.ndarray
        """
        np.random.seed(self.random_state)
        return np.random.randn(self.size)

    def _trend(self, sin_cos_noise_fact):
        """Get temporal trends.

        Parameters
        ----------
        sin_cos_noise_fact : Tuple of float, the weights of sine, cosine and standard gaussian distributed noise

        Returns
        -------
        numpy.ndarray
        """
        np.random.seed(self.random_state)
        return DataGenerator().trigonometry_ds(size=self.size, random_state=self.random_state,
                                               sin_cos_noise_fact=sin_cos_noise_fact)

    def event(self):
        """Get event data
        results = 10 * seasonal + 8 * trend + noise

        Returns
        -------
        numpy.ndarray
        """
        sin_cos_noise_facts = [np.random.choice(self.sin_cos_noise_fact, 3) for i in range(self.stacking_level)]
        if self.seasons is not None:
            x = self._seasonal() * 10 + self._trend(sin_cos_noise_facts[0]) * 8 + self._noise()

            if self.stacking_level > 1:
                for i in sin_cos_noise_facts[1:]:
                    x += self._trend(i) * 8
        else:
            x = self._trend(sin_cos_noise_facts[0]) * 8 + self._noise()

            if self.stacking_level > 1:
                for i in sin_cos_noise_facts[1:]:
                    x += self._trend(i) * 8
        return x



import torch
from torch import nn
from torch.nn import functional as F


class GaussianNoise1d(nn.Module):
    """Adds Gaussian noise to the input tensor.

    Parameters
    ----------
    level: float, the scaling multiple of gaussian noise

    Returns
    -------
    torch.Tensor
    """

    def __init__(self, level=0.1):
        super(GaussianNoise1d, self).__init__()
        self.level = level

    def forward(self, x):
        return x + torch.mul(torch.randn_like(x, device=x.device), 1 / torch.median(x) * self.level)


class Time2Vec(nn.Module):
    """Time2Vec module.

    Parameters
    ----------
    in_features: int, input-tensor last dimension shape

    Returns
    -------
    None
    """

    def __init__(self, in_features):
        super(Time2Vec, self).__init__()

        self.W = nn.Parameter(torch.randn(in_features, in_features))
        self.P = nn.Parameter(torch.randn(in_features))
        self.w = nn.Parameter(torch.randn(1))
        self.p = nn.Parameter(torch.randn(1))
        self.sine_w = nn.Parameter(torch.randn(1))
        self.cosine_w = nn.Parameter(torch.randn(1))

    def forward(self, x):
        original = self.w * x + self.p
        x = torch.matmul(x, self.W)
        for i in range(x.shape[0]):
            x[i, :] = self.sine_w * torch.sin(torch.squeeze(x[i]) + self.P) + \
                      self.cosine_w * torch.cos(torch.squeeze(x[i]) + self.P)

        return torch.concat((original, x), dim=-1)  # last dimension shape (, 2 * in_features)


class GAU(nn.Module):
    def __init__(
            self,
            in_features,
            query_key_dim=128,
            expansion_factor=2.,
            skip_connect=True,
            dropout=0.,
    ):
        super(GAU, self).__init__()
        hidden_dim = int(expansion_factor) * in_features

        self.norm = nn.LayerNorm(in_features)
        self.dropout = nn.Dropout(dropout)

        self.to_hidden = nn.Sequential(
            nn.Linear(in_features, hidden_dim * 2),
            nn.SiLU()
        )

        self.to_qk = nn.Sequential(
            nn.Linear(in_features, query_key_dim),
            nn.SiLU()
        )

        self.gamma = nn.Parameter(torch.ones(2, query_key_dim))
        self.beta = nn.Parameter(torch.zeros(2, query_key_dim))
        nn.init.normal_(self.gamma, std=0.02)

        self.to_out = nn.Sequential(
            nn.Linear(hidden_dim, in_features),
            nn.Dropout(dropout)
        )

        self.add_residual = skip_connect

    def forward(self, x):
        seq_len = x.shape[-2]

        normed_x = self.norm(x)  # (bs,seq_len,dim)
        v, gate = self.to_hidden(normed_x).chunk(2, dim=-1)  # (bs,seq_len,seq_len)

        Z = self.to_qk(normed_x)  # (bs,seq_len,query_key_dim)

        QK = torch.einsum('... d, h d -> ... h d', Z, self.gamma) + self.beta
        q, k = QK.unbind(dim=-2)

        if x.ndim == 2:
            sim = torch.einsum('i d, j d -> i j', q, k) / seq_len
        else:
            sim = torch.einsum('b i d, b j d -> b i j', q, k) / seq_len

        A = F.relu(sim) ** 2
        A = self.dropout(A)

        if x.ndim == 2:
            V = torch.einsum('i j, j d -> i d', A, v)
        else:
            V = torch.einsum('b i j, b j d -> b i d', A, v)

        V = V * gate

        out = self.to_out(V)

        if self.add_residual:
            out = out + x

        return out



from functools import update_wrapper
import inspect


class _AvailableIfDescriptor:
    def __init__(self, fn, check, attribute_name):
        self.fn = fn
        self.check = check
        self.attribute_name = attribute_name

        # update the docstring of the descriptor
        update_wrapper(self, fn)

    def __get__(self, obj, owner=None):
        attr_err = AttributeError(
            f"This {repr(owner.__name__)} has no attribute {repr(self.attribute_name)}"
        )
        if obj is not None:
            # delegate only on instances, not the classes.
            # this is to allow access to the docstrings.
            if not self.check(obj):
                raise attr_err

            # lambda, but not partial, allows help() to work with update_wrapper
            out = lambda *args, **kwargs: self.fn(obj, *args, **kwargs)  # noqa
        else:

            def fn(*args, **kwargs):
                if not self.check(args[0]):
                    raise attr_err
                return self.fn(*args, **kwargs)

            # This makes it possible to use the decorated method as an unbound method,
            # for instance when monkeypatching.
            out = lambda *args, **kwargs: fn(*args, **kwargs)  # noqa
        # update the docstring of the returned function
        update_wrapper(out, self.fn)
        return out


def available_if(check):
    return lambda fn: _AvailableIfDescriptor(fn, check, attribute_name=fn.__name__)


def func_has_params(func, param):
    sig = inspect.signature(func)
    params = sig.parameters.keys()
    if param in params:
        return True
    return False




from torch import nn


class SeriesRecombinationLayer(nn.Module):
    def __init__(self, in_shapes, out_features=128, rnn_layers=1, dropout=0.):
        assert isinstance(in_shapes, tuple) and len(in_shapes) == 2
        super(SeriesRecombinationLayer, self).__init__()
        (self.rows, self.cols), self.out_features = in_shapes, out_features

        self.encoder_rnn = nn.GRU(self.rows, out_features, batch_first=True, num_layers=rnn_layers,
                                  dropout=dropout if rnn_layers > 1 else 0.)
        self.decoder_rnn = nn.GRU(out_features, out_features, batch_first=True, num_layers=rnn_layers,
                                  dropout=dropout if rnn_layers > 1 else 0.)

        self.out = nn.Linear(self.cols * out_features, out_features)

    def forward(self, x):
        assert x.ndim == 3

        x = x.permute((0, 2, 1))
        x, h = self.encoder_rnn(x)
        x, h = self.decoder_rnn(x, h)
        x = x.permute((0, 2, 1))

        return self.out(x.reshape((-1, self.cols * self.out_features)))  # (batch_size, out_features)




import numpy as np
from spinesUtils.feature_tools import vars_threshold, variation_threshold

from spinesTS.metrics import r2_score


class ForecastingMixin:
    def extend_predict(self, x, n):
        """Extrapolation prediction.

        Parameters
        ----------
        x: to_predict data, must be 2 dims data
        n: predict steps, must be int

        Returns
        -------
        np.ndarray, which has 2 dims

        """
        assert isinstance(n, int)
        assert x.ndim == 2

        current_res = self.predict(x)

        if n is None:
            return current_res
        elif n <= current_res.shape[1]:
            return current_res[:, :n]
        else:
            res = [current_res]
            for i in range((n//res[0].shape[1])+1):
                current_res = self.predict(x)
                res.append(current_res)
                x = np.concatenate((x[:, current_res.shape[1]:], current_res), axis=-1)

            res = np.concatenate(res, axis=-1)[:, :n]

            return res

    def score(self, x, y, eval2d=True):
        if eval2d:
            return r2_score(y.T, self.predict(x).T)
        else:
            return r2_score(y, self.predict(x))


class TableFeatureGenerateMixin:
    """Table Feature Generate Mixin class"""
    def features_filter(self, x):
        to_remove_col_names = list(set(vars_threshold(x) + variation_threshold(x)))
        if len(to_remove_col_names) > 0:
            x.drop(columns=to_remove_col_names, inplace=True)

        return x



from typing import Any, Union

import torch
from torch import nn

from spinesTS.base import TorchModelMixin, ForecastingMixin


def get_even_blocks(in_features, window_size):
    """
    将输入数据分割成多个块，每个块的大小为window_size
    :param in_features:
    :return:
    """

    blocks = in_features // window_size + int(in_features % window_size != 0)

    while (blocks % 2 != 0 or window_size % 2 != 0) and window_size >= 6:
        window_size -= 1
        blocks = in_features // window_size + int(in_features % window_size != 0)

    return blocks, window_size


class SegmentationBlock(nn.Module):
    """将输入数据分割成多个块，每个块的大小为window_size"""

    def __init__(self, in_features, window_size=12):
        super(SegmentationBlock, self).__init__()

        self.blocks, self.window_size = get_even_blocks(in_features, window_size)

        if self.blocks % 2 != 0:
            self.blocks += 1

        self.encoders = nn.ModuleList([
            nn.Sequential(
                nn.Linear(self.window_size, self.window_size),
                nn.SELU()
            )
            for i in range(self.blocks)
        ])

    def forward(self, x):
        assert x.ndim == 2

        if self.blocks == 1:
            xs = [x]
        else:
            xs = list(torch.split(x, self.window_size, dim=1))
            xs[-1] = torch.concat([xs[-1], torch.zeros(xs[-1].shape[0],
                                                       self.window_size - xs[-1].shape[1])], dim=1)

        res = []
        for i in range(self.blocks):
            res.append(self.encoders[i](xs[i]))

        return torch.stack(res, 1)  # [batch_size, blocks, window_size]


class SegRNNBlock(nn.Module):
    def __init__(self, in_features, out_features):
        super(SegRNNBlock, self).__init__()

        if in_features < 24:
            raise ValueError("The number of input features must be equal or greater than 24.")

        self.splitter = SegmentationBlock(in_features=in_features, window_size=12)
        window_size = self.splitter.window_size

        self.encoder_rnn = nn.GRU(window_size, window_size, num_layers=1,
                                  bias=True, bidirectional=False, batch_first=True)

        self.decoder_rnn = nn.GRU(window_size, window_size, num_layers=1,
                                  bias=True, bidirectional=False, batch_first=True)

        out_window_size = out_features // self.splitter.blocks + int(out_features % self.splitter.blocks > 0)

        self.decoders = nn.ModuleList([
            nn.Sequential(
                nn.Linear(window_size, 1024),
                nn.SELU(),
                nn.Linear(1024, out_window_size)
            )
            for i in range(self.splitter.blocks)
        ])

        self.out_features = out_features

    def forward(self, x):
        assert x.shape[1] % 2 == 0, "Only even numbers of columns are accepted."

        x = self.splitter(x)  # [batch_size, blocks, window_size]

        _, h = self.encoder_rnn(x)

        output, _ = self.decoder_rnn(x, h)

        res = []
        for i in range(self.splitter.blocks):
            res.append(self.decoders[i](output[:, i, :]))

        res = torch.concat(res, dim=1)

        return res[:, :self.out_features]


class SegRNN(TorchModelMixin, ForecastingMixin):
    """长序列预测模型

    References
    ----------
    LIN S. SegRNN: Segment Recurrent Neural Network for Long-Term Time Series Forecasting[J]. 2023.
    arXiv preprint arXiv:2308.11200
    """

    def __init__(self,
                 in_features: int,
                 out_features: int,
                 loss_fn='mae',
                 learning_rate: float = 0.001,
                 random_seed: int = 42,
                 device='cpu'
                 ) -> None:
        super(SegRNN, self).__init__(random_seed, device, loss_fn=loss_fn)
        self.in_features, self.out_features = in_features, out_features
        self.learning_rate = learning_rate
        self.model, self.loss_fn, self.optimizer = self.call()

    def call(self) -> tuple:
        model = SegRNNBlock(
            in_features=self.in_features,
            out_features=self.out_features
        )
        loss_fn = self.loss_fn
        optimizer = torch.optim.AdamW(model.parameters(), lr=self.learning_rate)
        return model, loss_fn, optimizer

    def fit(self,
            X_train: Any,
            y_train: Any,
            epochs: int = 1000,
            batch_size: Union[str, int] = 'auto',
            eval_set: Any = None,
            monitor: str = 'val_loss',
            min_delta: int = 0,
            patience: int = 10,
            lr_scheduler: Union[str, None] = 'ReduceLROnPlateau',
            lr_scheduler_patience: int = 10,
            lr_factor: float = 0.7,
            restore_best_weights: bool = True,
            verbose: bool = True,
            **kwargs: Any) -> Any:
        return super().fit(X_train, y_train, epochs, batch_size, eval_set, loss_type='down', metrics_name='mae',
                           monitor=monitor, lr_scheduler=lr_scheduler,
                           lr_scheduler_patience=lr_scheduler_patience,
                           lr_factor=lr_factor,
                           min_delta=min_delta, patience=patience, restore_best_weights=restore_best_weights,
                           verbose=verbose, **kwargs)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from typing import *
from matplotlib.figure import Figure


def plot1d(*S: Sequence, x: Optional[Sequence] = None,
           title=None,
           labels: Optional[Sequence] = None,
           figsize: Sequence = (12, 8),
           legend: bool = True, linestyle='-', colors=None,
           legend_loc='best'
           ) -> Figure:
    """Plot series, accept one dim array-like sequence

    Parameters
    ----------
    S: array-like series, which to plot.
    x: None or array-like series, x-axis.
    title: figure title
    labels: None or List[str] or str-type, figure-label.
    figsize: tuple of integers, figure size
    legend: bool, whether to show the figure legend
    linestyle: line style, str-like
    legend_loc: str or pair of floats, default: rcParams["legend.loc"] (default: 'best'), it can be
        'upper left', 'upper right', 'lower left', 'lower right',
        'upper center', 'lower center', 'center left', 'center right',
        'center', 'best'

    Returns
    -------
    matplotlib.figure.Figure
    """

    if isinstance(colors, str) or colors is None:
        colors = [colors for i in range(len(S))]
    else:
        assert isinstance(colors, (list, tuple))
        assert len(colors) == len(S)

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    ax.set_title(title)
    seq_len = len(S)

    for i in range(seq_len):
        if labels is not None:
            if isinstance(labels, str):
                label = labels
            else:
                assert len(labels) == seq_len, "labels length must be equals to sequence length when labels is a list"
                label = labels[i]
        else:
            label = 'label_' + str(i + 1)

        if x is None:
            ax.plot(S[i], label=label, linestyle=linestyle, color=colors[i])
        else:
            ax.plot(x, S[i], label=label, linestyle=linestyle, color=colors[i])

        if legend:
            plt.legend(loc=legend_loc)

    fig.tight_layout()

    return ax


def plot2d(*S: Sequence, x: Optional[Sequence] = None,
           title=None, linestyle='-',
           labels: Optional[Sequence[str]] = None,
           eval_slices: str = '[-1]',
           figsize: Tuple[int, int] = (10, 6),
           legend: bool = True,
           subplots_shape: Tuple[int, int] = (-1, 5), colors=None,
           legend_loc='best'
           ):
    """Plot series, accept two dim array-like sequence

    Parameters
    ----------
    S: array-like series, which to plot.
    x: None or array-like series, x-axis.
    labels: None or List[str], figure-label.
    eval_slices: str,
    figsize: tuple of integers, figure size
    legend: bool, whether to show the figure legend
    subplots_shape: tuple
    title: figure title, str-like
    linestyle: line style, str-like
    colors: line colors, None or str-like or list, or tuple
    legend_loc: str or pair of floats, default: rcParams["legend.loc"] (default: 'best'), it can be
        'upper left', 'upper right', 'lower left', 'lower right',
        'upper center', 'lower center', 'center left', 'center right',
        'center', 'best'

    Returns
    -------
    matplotlib.figure.Figure
    """
    assert np.sum([np.ndim(i) == 2 for i in S]) == len(S)
    assert isinstance(eval_slices, str)

    if isinstance(colors, str) or colors is None:
        colors = [colors for i in range(len(S))]
    else:
        assert isinstance(colors, (list, tuple))
        assert len(colors) == len(S)

    fig = plt.figure(facecolor='w', figsize=figsize)

    _to_plot_array = []
    for _s in range(len(S)):
        _v = eval('S[_s]' + eval_slices)
        if not isinstance(_v[0], (list, tuple, np.ndarray, pd.Series)):
            _v = [_v]
        _to_plot_array.append(_v)

    num = max([len(_) for _ in _to_plot_array])

    cols = subplots_shape[-1] if num >= 5 else 1
    rows = num if num < 5 else (num // subplots_shape[-1] + 1 if num % subplots_shape[-1] > 0
                                else num // subplots_shape[-1])

    for j in range(len(_to_plot_array)):  # figures
        for i in range(1, num + 1):
            ax = plt.subplot(rows, cols, i)
            if title is not None:
                ax.set_title(title+f' {i-1}')

            if labels is None:
                label = f'label_' + str(j + 1)
            else:
                label = labels[j]
            if x is not None:
                ax.plot(x, _to_plot_array[j][i - 1], label=label, linestyle=linestyle, color=colors[j])
            else:
                ax.plot(_to_plot_array[j][i - 1], label=label, linestyle=linestyle, color=colors[j])
            ax.grid(True, which='major', c='gray', ls='-', lw=1, alpha=0.2)
            if legend:
                ax.legend(loc=legend_loc)

    fig.tight_layout()
    return fig



import re
import pandas as pd
import os
from tabulate import tabulate

from ..frame import DataTS

FILE_PATH = os.path.dirname(__file__)


def _call_name(built_in_func, name):
    """Wrapper of built-in datasets

    Parameters
    ----------
    built_in_func : BuiltInSeriesData class
    name : dataset's name

    Returns
    -------
    wrapped func
    """
    assert name is not None

    def wrap():
        return built_in_func[name]

    return wrap


class BuiltInSeriesData:
    """Load the built-in data

    Parameters
    ----------
    print_file_list : bool, whether to print the exists file name list

    Returns
    -------
    None
    """

    def __init__(self, print_file_list=True):
        self.file_list = sorted(os.listdir(os.path.join(FILE_PATH, './built-in-datasets/')))
        if print_file_list:
            print(self)

    def _load_data(self, fp):
        self._FILEPATH = os.path.join(FILE_PATH, './built-in-datasets/', fp)
        if not os.path.exists(self._FILEPATH):
            self._FILEPATH = fp
        assert os.path.exists(self._FILEPATH), f'No such file or directory: {self._FILEPATH}'

        return DataTS(pd.read_csv(self._FILEPATH, sep=','), name='.'.join(fp.split('.')[:-1]))

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._load_data(self.file_list[item])

        elif isinstance(item, str):
            if not item.endswith('.csv'):
                item = item + '.csv'
            return self._load_data(self.file_list[self.file_list.index(item)])
        else:
            raise KeyError(f"invalid key: {item}")

    def __len__(self):
        return len(self.file_list)

    @property
    def names(self):
        """Returns the built-in series data names-list."""
        return [''.join(i.split('.')[:-1]) for i in self.file_list]

    def __str__(self):
        table = []
        for i in range(len(self.file_list)):
            _ = [re.split('\.', self.file_list[i])[0].strip(),
                 ', '.join(self[i].columns.tolist())]
            table.append(_)

        return tabulate(table, headers=["ds name", "columns"], showindex="always",
                        tablefmt="pretty", colalign=("right", "left", "left"))

    def __repr__(self):
        return self.__str__()


LoadElectricDataSets = _call_name(BuiltInSeriesData(print_file_list=False), 'Electric_Production')
LoadMessagesSentDataSets = _call_name(BuiltInSeriesData(print_file_list=False), 'Messages_Sent')
LoadMessagesSentHourDataSets = _call_name(BuiltInSeriesData(print_file_list=False), 'Messages_Sent_Hour')
LoadWebSales = _call_name(BuiltInSeriesData(print_file_list=False), 'Web_Sales')
LoadSupermarketIncoming = _call_name(BuiltInSeriesData(print_file_list=False), 'Supermarket_Incoming')



from joblib import dump
from joblib import load


def save_model(model_obj, file_name):
    """Use pickle to save python models.

    Parameters
    ----------
    model_obj: model object
    file_name: str, file-path-like

    Returns
    -------
    None
    """
    dump(model_obj, file_name)


def load_model(file_name):
    """Use pickle to load python models.

    Parameters
    ----------
    file_name: str, file-path-like

    Returns
    -------
    None
    """
    return load(file_name)



from ._decompose_layers import (
    Hierarchical1d,
    DimensionConv1d,
    MoveAvg,
    DifferentialLayer
)
from ._concat_layers import RecurseResBlock
from ._enhance_layers import (
    GaussianNoise1d,
    Time2Vec,
    GAU
)
from ._multi_features import SeriesRecombinationLayer
from ._position_encoder import PositionalEncoding



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
  long_description = fh.read()

setup(
    name='spinesTS',
    version="0.3.13",
    description='spinesTS, a powerful timeseries toolsets.',
    keywords='machine learning',
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    url='https://github.com/BirchKwok/spinesTS',
    author='Birch Kwok',
    author_email='birchkwok@gmail.com',
    install_requires=[
        'scikit-learn>=1.0.2',
        'torch>=2.1.0',
        'scipy>=1.7.0',
        'numpy>=1.17.0',
        'pandas>=2.0.0',
        'tabulate>=0.8',
        'matplotlib>=3.5.1',
        'spinesUtils>=0.3.5'
    ],
    zip_safe=False,
    include_package_data=True
)



import pandas as pd


def moving_average(x, window_size=3):
    assert window_size > 1
    if x.ndim == 1:
        x = pd.Series(x)
        return x.rolling(window_size).mean().values[window_size-1:]
    elif x.ndim == 2:
        x = pd.DataFrame(x.T)
        return x.rolling(window_size).mean().values.T[:, window_size-1:]
    else:
        raise ValueError("x must be one dim or two dim sequence.")
    